from __future__ import annotations

import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, TextIO

from .._colors import get_colors
from ..linter._extractors import get_contracts
from ._base import Command
from ._common import get_paths


try:
    import deal_solver
except ImportError:
    deal_solver = None  # type: ignore[assignment]
try:
    from deal_solver import Theorem
except ImportError:
    Theorem = object  # type: ignore[misc,assignment]

if TYPE_CHECKING:
    import astroid


TEMPLATE_MOD = '{blue}{name}{end}'
TEMPLATE_FUN = '  {magenta}{name}{end}'
TEMPLATE_CON = '    {color}{p.conclusion.value}{end} {p}'


class DealTheorem(Theorem):
    @staticmethod
    def get_contracts(func: astroid.FunctionDef) -> Iterator[deal_solver.Contract]:
        for contract in get_contracts(func):
            yield deal_solver.Contract(
                name=contract.name,
                args=contract.args,  # type: ignore[arg-type]
            )


def run_solver(
    path: Path,
    stream: TextIO,
    show_skipped: bool,
    colors: dict[str, str],
) -> int:
    file_name_shown = False
    text = path.read_text()
    theorems = DealTheorem.from_text(text)
    failed_count = 0
    for theorem in theorems:
        if theorem.name.startswith('test_'):
            continue

        proof = theorem.prove()
        assert proof.conclusion is not None
        if proof.conclusion == deal_solver.Conclusion.SKIP and not show_skipped:
            continue

        if not file_name_shown:
            line = TEMPLATE_MOD.format(name=path, **colors)
            print(line, file=stream)
            file_name_shown = True

        line = TEMPLATE_FUN.format(name=theorem.name, **colors)
        print(line, file=stream)
        line = TEMPLATE_CON.format(p=proof, color=colors[proof.color], **colors)
        print(line, file=stream)
        failed_count += proof.conclusion == deal_solver.Conclusion.FAIL
    return failed_count


class ProveCommand(Command):
    """Verify correctness of code.

    ```bash
    python3 -m deal prove project/
    ```

    Options:

    + `--skipped`: show skipped functions.
    + `--nocolor`: disable colored output.

    Exit code is equal to the failed theorems count.
    See [Formal Verification][verification] documentation for more information.

    [verification]: https://deal.readthedocs.io/basic/verification.html
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('--skipped', action='store_true', help='show skipped')
        parser.add_argument('--nocolor', action='store_true', help='colorless output')
        parser.add_argument('paths', nargs='+')

    def __call__(self, args) -> int:
        if deal_solver is None:  # pragma: no cover
            self.print('deal-solver is required but not installed')
            self.print(f'{sys.executable} -m pip install deal-solver')
            return 1
        colors = get_colors(args)
        failed = 0
        for arg in args.paths:
            for path in get_paths(Path(arg)):
                failed += run_solver(
                    path=path,
                    stream=self.stream,
                    show_skipped=args.skipped,
                    colors=colors,
                )
        return failed



from pathlib import Path

import yaml

from ._core import run_test


def main():
    path = Path(__file__).parent / 'tests.yaml'
    for group in yaml.safe_load(path.read_text()):
        print(group['name'])
        for item in group['items']:
            run_test(**item)



import pytest
import urllib3

import deal

from .helpers import run_sync


def test_raises_exception():

    @deal.has()
    def func(do):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')

    func(False)
    with pytest.raises(deal.OfflineContractError):
        func(True)


def test_raises_specified_exception():

    @deal.has(exception=KeyError)
    def func(do):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')

    func(False)
    with pytest.raises(KeyError):
        func(True)


def test_allow_network():

    @deal.has('network')
    def func(do):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')

    func(False)
    func(True)


def test_decorating_async_function():
    @deal.has()
    async def func(do) -> int:
        if not do:
            return 1
        http = urllib3.PoolManager()
        http.request('GET', 'http://httpbin.org/robots.txt')

    assert run_sync(func(False)) == 1
    with pytest.raises(deal.OfflineContractError):
        run_sync(func(True))


def test_decorating_generator():
    @deal.has()
    def func(do):
        if not do:
            yield 1
            return
        http = urllib3.PoolManager()
        http.request('GET', 'http://httpbin.org/robots.txt')

    assert list(func(False)) == [1]
    with pytest.raises(deal.OfflineContractError):
        list(func(True))



import json
from pathlib import Path

from deal._cli import main


TEXT = """
import deal

def f(x):
    raise PatheticError
"""


def test_stub_command(tmp_path: Path):
    path = (tmp_path / 'example.py')
    path.write_text(TEXT)
    result = main(['stub', '--iterations', '1', str(path)])
    assert result == 0
    content = (tmp_path / 'example.json').read_text()
    assert json.loads(content) == {'f': {'raises': ['PatheticError']}}



"""
**Deal** is a Python library for [design by contract][wiki] (DbC) programming.
See [documentation] for more details.

[wiki]: https://en.wikipedia.org/wiki/Design_by_contract
[documentation]: https://deal.readthedocs.io/index.html
"""
from . import introspection
from ._exceptions import (
    ContractError, ExampleContractError, InvContractError, MarkerError,
    NoMatchError, OfflineContractError, PostContractError, PreContractError,
    RaisesContractError, ReasonContractError, SilentContractError,
)
from ._imports import activate, module_load
from ._runtime import (
    catch, chain, dispatch, ensure, example, has, implies,
    inherit, inv, post, pre, pure, raises, reason, safe,
)
from ._schemes import Scheme
from ._sphinx import autodoc
from ._state import disable, enable, reset
from ._testing import TestCase, cases


__title__ = 'deal'
__version__ = '4.23.1'
__author__ = 'Gram (@orsinium)'
__license__ = 'MIT'
__all__ = [
    'autodoc',
    'cases',
    'introspection',
    'Scheme',
    'TestCase',

    # state
    'disable',
    'enable',
    'reset',

    # decorators
    'chain',
    'dispatch',
    'ensure',
    'example',
    'has',
    'inherit',
    'inv',
    'post',
    'pre',
    'raises',
    'reason',

    # aliases
    'catch',
    'implies',
    'pure',
    'safe',

    # module level
    'module_load',
    'activate',

    # exceptions
    'ContractError',
    'ExampleContractError',
    'InvContractError',
    'MarkerError',
    'NoMatchError',
    'OfflineContractError',
    'PostContractError',
    'PreContractError',
    'RaisesContractError',
    'ReasonContractError',
    'SilentContractError',
]



from __future__ import annotations

import ast
from pathlib import Path
from typing import Iterator, NamedTuple

import astroid

from ._contract import Category, Contract
from ._extractors import get_contracts, get_definitions


class Func(NamedTuple):
    name: str
    body: list
    contracts: list[Contract]
    node: ast.FunctionDef | astroid.FunctionDef

    @property
    def line(self) -> int:
        return self.node.lineno

    @property
    def col(self) -> int:
        return self.node.col_offset

    @property
    def has_self(self) -> bool:
        """Check if the first function argument is `self`.
        """
        args = self.node.args.args
        if not args:
            return False
        arg = args[0]
        if isinstance(arg, ast.arg):
            return arg.arg == 'self'
        return arg.name == 'self'

    @classmethod
    def from_path(cls, path: Path) -> list[Func]:
        text = path.read_text()
        tree = astroid.parse(code=text, path=str(path))
        return cls.from_astroid(tree)

    @classmethod
    def from_text(cls, text: str) -> list[Func]:
        tree = astroid.parse(text)
        return cls.from_astroid(tree)

    @classmethod
    def from_ast(cls, tree: ast.Module) -> list[Func]:
        funcs = []
        definitions = get_definitions(tree=tree)
        for expr in cls._get_funcs_ast(tree):
            contracts = []
            for cinfo in get_contracts(expr):
                contract = Contract(
                    args=cinfo.args,
                    func_args=expr.args,
                    category=Category(cinfo.name),
                    context=definitions,
                    line=cinfo.line,
                )
                contracts.append(contract)
            funcs.append(cls(
                name=expr.name,
                body=expr.body,
                contracts=contracts,
                node=expr,
            ))
        return funcs

    @classmethod
    def _get_funcs_ast(cls, target: ast.AST) -> Iterator[ast.FunctionDef]:
        if isinstance(target, ast.Module):
            for stmt in target.body:
                yield from cls._get_funcs_ast(stmt)
        elif isinstance(target, ast.FunctionDef):
            yield target
        elif isinstance(target, ast.ClassDef):
            for stmt in target.body:
                if isinstance(stmt, ast.FunctionDef):
                    yield stmt

    @classmethod
    def from_astroid(cls, tree: astroid.Module) -> list[Func]:
        funcs = []
        definitions = get_definitions(tree=tree)
        for expr in cls._get_funcs_astroid(tree):
            # make signature
            code = f'def f({expr.args.as_string()}):0'
            func_args = ast.parse(code).body[0].args  # type: ignore

            # collect contracts
            contracts = []
            for cinfo in get_contracts(expr):
                contract = Contract(
                    args=cinfo.args,
                    func_args=func_args,
                    category=Category(cinfo.name),
                    context=definitions,
                    line=cinfo.line,
                )
                contracts.append(contract)
            assert expr.lineno is not None
            funcs.append(cls(
                name=expr.name,
                body=expr.body,
                contracts=contracts,
                node=expr,
            ))
        return funcs

    @classmethod
    def _get_funcs_astroid(cls, target: astroid.NodeNG) -> Iterator[astroid.FunctionDef]:
        if isinstance(target, astroid.Module):
            for stmt in target.body:
                yield from cls._get_funcs_astroid(stmt)
        elif isinstance(target, astroid.FunctionDef):
            yield target
        elif isinstance(target, astroid.ClassDef):
            for stmt in target.body:
                if isinstance(stmt, astroid.FunctionDef):
                    yield stmt

    def has_contract(self, *categories: Category) -> bool:
        """Check if the function has a contract from any of the given categories.
        """
        for contract in self.contracts:
            if contract.category in categories:
                return True
        return False

    def __repr__(self) -> str:
        cats = ', '.join(contract.category.value for contract in self.contracts)
        return f'{type(self).__name__}({cats})'



import ast
import sys
from pathlib import Path
from textwrap import dedent
from typing import Dict

import astroid
import docstring_parser
import pytest

from deal.linter._extractors import get_exceptions
from deal.linter._extractors.exceptions import _excs_from_doc


@pytest.mark.parametrize('text, expected', [
    ('raise BaseException', (BaseException, )),
    ('raise ValueError', (ValueError, )),
    ('raise UnknownError', ('UnknownError', )),
    ('raise ValueError("lol")', (ValueError, )),
    ('raise unknown()', ()),
    ('raise 1 + 2', ()),
    ('raise', ()),
    ('assert False', (AssertionError, )),
    ('12 / 0', (ZeroDivisionError, )),
    ('12 + 0', ()),
    ('exit()', (SystemExit, )),
    ('exit(13)', (SystemExit, )),
    ('sys.exit(13)', (SystemExit, )),
    ('something.exit(13)', ()),

    # try-except
    ('try:\n raise AError\nexcept Exception:\n pass', ()),
    ('try:\n raise AError\nexcept AError:\n raise BError', ('BError', )),
    ('try:\n pass\nfinally:\n raise KeyError', (KeyError, )),

    ('if True: raise KeyError', (KeyError, )),
    ('for i in lst: raise KeyError', (KeyError, )),
])
def test_get_exceptions_simple(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    returns = tuple(r.value for r in get_exceptions(body=tree.body))
    assert returns == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    returns = tuple(r.value for r in get_exceptions(body=tree.body))
    assert returns == expected


def test_inference_simple():
    text = """
        def subf():
            raise ValueError  # explicit raise

        def subf2():
            c = 1 - 1
            1 / c   # implicit raise
            d = [1, 2, 3]
            1 / d   # resolved into not a constant
            a = b

        @deal.raises(KeyError)
        def f():
            a = 1
            a()         # resolved not in a function
            unknown()   # cannot resolve
            subf()      # resolve
            subf2()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == (ValueError, ZeroDivisionError)


def test_inference_assign():
    text = """
        def subf():
            raise Unknown

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ('Unknown', )


def test_inference_ok_uncalled():
    text = """
        def subf():
            raise ValueError

        @deal.raises(KeyError)
        def f():
            subf
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ()


def test_inference_subcalls():
    text = """
        def subf():
            raise ValueError

        def subf2():
            raise IndexError

        @deal.raises(KeyError)
        def f():
            other(subf(), b=subf2())
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == (ValueError, IndexError)


def test_resolve_doesnt_fail_for_simple_ast():
    text = """
        def subf():
            raise ValueError  # explicit raise

        @deal.raises(KeyError)
        def f():
            subf()
    """
    tree = ast.parse(dedent(text))
    print(ast.dump(tree))
    func = tree.body[-1]
    assert isinstance(func, ast.FunctionDef)
    tuple(get_exceptions(body=func.body))


def test_inference_subcontracts():
    text = """
        @deal.raises(SomeError)     # actual contract
        @deal.raises(1)             # ignore junk
        @deal.post(lambda _: 1)     # ignore other contracts
        def subf():
            return 1

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ('SomeError', )


def test_inference_doesnt_have_exceptions():
    text = """
        def subf():
            something()
            return 1

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == ()


@pytest.fixture
def remove_import():
    old_imports: Dict[str, Dict[str, object]] = dict()

    def _remove_import(what: str):
        old_imports[what] = dict()
        for mod_name, module in sys.modules.items():
            if not mod_name.startswith('deal.'):
                continue
            if not hasattr(module, what):
                continue
            old_imports[what][mod_name] = getattr(module, what)
            setattr(module, what, None)
    yield _remove_import

    for what, imports in old_imports.items():
        for mod_name, old_import in imports.items():
            module = sys.modules[mod_name]
            setattr(module, what, old_import)


@pytest.mark.parametrize('docstring', [
    # sphinx
    pytest.param(
        """Does not raise RuntimeError.

        :raises ValueError:
        :raises KeyError: some junk
        :returns: RuntimeError
        """,
        id='sphinx',
    ),
    # epydoc
    pytest.param(
        """Does not raise RuntimeError.

        @raise ValueError:
        @raise KeyError: some junk
        @raise: something
        @return: RuntimeError
        """,
        id='epydoc',
    ),
    # google
    pytest.param(
        """Does not raise RuntimeError.

        Raises:
            ValueError:
                some junk
            KeyError: some junk
        Returns:
            RuntimeError
        """,
        id='google',
    ),
    # numpy
    pytest.param(
        """Does not raise RuntimeError.

        Raises
        ------
        ValueError
                some junk
        KeyError

        Returns
        -------
        RuntimeError
        """,
        id='numpy',
    ),
])
@pytest.mark.parametrize('patch_third_party', [True, False])
def test_extract_from_docstring(docstring, patch_third_party, remove_import):
    if patch_third_party:
        remove_import('docstring_parser')
    text = """
        def subf():
            '''{docstring}'''
            something()
            return 1

        @deal.raises(KeyError)
        def f():
            b = subf()
    """
    text = dedent(text).format(docstring=docstring)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree))
    assert returns == (ValueError, KeyError)


def get_docstring_parser_tests():
    """Extract test cases from docstring_parser/tests.
    """
    root = Path(docstring_parser.__path__[0]) / 'tests'
    tests_found = 0
    for path in root.iterdir():
        if path.suffix != '.py':
            continue
        tree = astroid.parse(code=path.read_text(), path=str(path))
        assert isinstance(tree, astroid.Module)
        for func in tree.body:
            if not isinstance(func, astroid.FunctionDef):
                continue
            if not func.decorators:
                continue
            assert isinstance(func.decorators, astroid.Decorators)
            for dec in func.decorators.nodes:
                if not isinstance(dec, astroid.Call):
                    continue
                if dec.func.as_string() != 'pytest.mark.parametrize':
                    continue
                cases = dec.args[1]
                assert isinstance(cases, astroid.List)
                for case in cases.elts:
                    assert isinstance(case, astroid.Tuple)
                    given = case.elts[0]
                    if not isinstance(given, astroid.Const):
                        continue
                    yield pytest.param(given.value, id=f'{path.name}:{given.lineno}')
                    tests_found += 1
    assert tests_found > 10


@pytest.mark.parametrize('docstring', get_docstring_parser_tests())
def test_compare_with_docstring_parser(docstring, remove_import):
    """
    Run test cases from docstring_parser and run extractor with and without
    docstring_parser. The result must be the same.
    """
    theirs = list(_excs_from_doc(docstring))
    remove_import('docstring_parser')
    ours = list(_excs_from_doc(docstring))
    assert theirs == ours



from __future__ import annotations

from typing import Callable, Iterator, TypeVar

from .._runtime import Contracts, Inherit
from . import _wrappers
from ._wrappers import Contract, ValidatedContract


ATTR = '__deal_contract'
F = TypeVar('F', bound=Callable)


def unwrap(func: F) -> F:
    """Get the original wrapped function without deal contracts.
    """
    contracts = getattr(func, ATTR, None)
    if isinstance(contracts, Contracts):
        return contracts.func
    return func


def init_all(func: Callable) -> None:
    """Call .init() on all contracts.

    This is useful when you need to explicitly precache contracts.
    For example, when profiling the function.
    """
    for contract in get_contracts(func):
        if isinstance(contract, ValidatedContract):
            contract.init()


def get_contracts(func: Callable) -> Iterator[Contract]:
    while True:
        if isinstance(func, Inherit):
            func = func._patch()
        contracts = getattr(func, ATTR, None)
        if isinstance(contracts, Contracts):
            for validator in contracts.pres:
                yield _wrappers.Pre(validator)
            for validator in contracts.posts:
                yield _wrappers.Post(validator)
            for validator in contracts.ensures:
                yield _wrappers.Ensure(validator)
            for validator in contracts.raises:
                yield _wrappers.Raises(validator)
            for validator in contracts.reasons:
                yield _wrappers.Reason(validator)
            for validator in contracts.examples:
                yield _wrappers.Example(validator)
            if contracts.patcher:
                yield _wrappers.Has(contracts.patcher)

        if not hasattr(func, '__wrapped__'):
            return
        func = func.__wrapped__  # type: ignore[attr-defined]



import deal


def preserve_function_annotation() -> None:
    @deal.has('db', 'network')
    def f(a: int, b: int) -> str:
        return str(a + b)
    reveal_type(f)  # R: def (a: builtins.int, b: builtins.int) -> builtins.str



from __future__ import annotations

import re
import sys
import traceback
from argparse import ArgumentParser
from contextlib import contextmanager
from functools import update_wrapper
from importlib import import_module
from pathlib import Path
from textwrap import indent
from typing import Iterable, Iterator, TextIO, TypeVar

from .._colors import COLORS
from .._testing import TestCase, cases
from .._trace import TraceResult, format_lines, trace
from ..linter._contract import Category
from ..linter._extractors.pre import format_call_args
from ..linter._func import Func
from ._base import Command
from ._common import get_paths


try:
    import pygments
except ImportError:
    pygments = None
else:
    from pygments.formatters import TerminalFormatter
    from pygments.lexers import PythonTracebackLexer


T = TypeVar('T')

rex_exception = re.compile(r'deal\.(\w*ContractError)')


@contextmanager
def sys_path(path: Path):
    path = str(path)
    sys.path.insert(0, path)
    try:
        yield
    finally:
        if sys.path[0] == path:
            del sys.path[0]


def has_pure_contract(func: Func) -> bool:
    for contract in func.contracts:
        if contract.category == Category.PURE:
            return True
        if contract.category == Category.HAS and not contract.args:
            return True
    return False


def get_func_names(path: Path) -> Iterator[str]:
    for func in Func.from_path(path=path):
        if has_pure_contract(func=func):
            yield func.name


def color_exception(text: str) -> str:
    text = rex_exception.sub(r'\1', text)
    if pygments is None:  # pragma: no cover
        return text
    return pygments.highlight(
        code=text,
        lexer=PythonTracebackLexer(),
        formatter=TerminalFormatter(),
    )


def format_exception() -> str:
    lines = traceback.format_exception(*sys.exc_info())
    text = color_exception(''.join(lines))
    text = indent(text=text, prefix='    ')
    return text.rstrip()


def fast_iterator(items: Iterable[T]) -> Iterator[T]:
    """
    Iterate over `iterator` disabling tracer on every iteration step.
    This is a trick to avoid using our coverage tracer when calling hypothesis machinery.
    Without it, testing is about 3 times slower.
    """
    iterator = iter(items)
    default_trace = sys.gettrace()
    while True:  # pragma: no cover
        sys.settrace(None)
        try:
            case = next(iterator)
        except StopIteration:
            return
        finally:
            sys.settrace(default_trace)
        yield case


def run_cases(
    cases: Iterator[TestCase],
    func_name: str,
    stream: TextIO,
    colors: dict[str, str],
) -> bool:
    print('  {blue}running {name}{end}'.format(name=func_name, **colors), file=stream)
    for case in cases:
        try:
            case()
        except Exception:
            line = '    {yellow}{name}({args}){end}'.format(
                name=func_name,
                args=format_call_args(args=case.args, kwargs=case.kwargs),
                **colors,
            )
            print(line, file=stream)
            print(format_exception(), file=stream)
            return False
    return True


def format_coverage(tresult: TraceResult, colors: dict[str, str]) -> str:
    cov = tresult.coverage
    if cov >= 85:
        color = colors['green']
    elif cov >= 50:
        color = colors['yellow']
    else:
        color = colors['red']
    tmpl = '    coverage {color}{cov}%{end}'
    missing = format_lines(
        statements=tresult.all_lines,
        lines=tresult.all_lines - tresult.covered_lines,
    )
    if cov != 0 and cov != 100 and len(missing) <= 60:
        tmpl += ' (missing {missing})'
    line = tmpl.format(
        cov=cov,
        color=color,
        missing=missing,
        **colors,
    )
    return line


class TestCommand(Command):
    """Generate and run tests against pure functions.

    ```bash
    python3 -m deal test project/
    ```

    Function must be decorated by one of the following to be run:

    + `@deal.pure`
    + `@deal.has()` (without arguments)

    Options:

    + `--count`: how many input values combinations should be checked.

    Exit code is equal to count of failed test cases.
    See [tests][tests] documentation for more details.

    [tests]: https://deal.readthedocs.io/basic/tests.html
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('--count', type=int, default=50)
        parser.add_argument('paths', nargs='+')

    def __call__(self, args) -> int:
        failed = 0
        for arg in args.paths:
            for path in get_paths(Path(arg)):
                failed += self.run_tests(
                    path=Path(path),
                    count=args.count,
                )
        return failed

    def run_tests(self, path: Path, count: int) -> int:
        names = list(get_func_names(path=path))
        if not names:
            return 0
        self.print('{magenta}running {path}{end}'.format(path=path, **COLORS))
        module_name = '.'.join(path.relative_to(self.root).with_suffix('').parts)
        with sys_path(path=self.root):
            module = import_module(module_name)
        failed = 0
        for func_name in names:  # pragma: no cover
            func = getattr(module, func_name)
            # set `__wrapped__` attr so `trace` can find the original function.
            runner = update_wrapper(wrapper=run_cases, wrapped=func)
            tresult = trace(
                func=runner,
                cases=fast_iterator(cases(func=func, count=count)),
                func_name=func_name,
                stream=self.stream,
                colors=COLORS,
            )
            if tresult.func_result and tresult.all_lines:
                text = format_coverage(tresult=tresult, colors=COLORS)
                self.print(text)
            else:
                failed += 1
        return failed   # pragma: no cover



from io import StringIO
from pathlib import Path
from textwrap import dedent

from deal._cli import main


def test_incorrect(tmp_path: Path):
    text = """
        import deal

        @deal.post(lambda x: x > 0)
        def func(x: int):
            return -1
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = main(['prove', str(path), '--nocolor'], stream=stream)

    stream.seek(0)
    captured = stream.read()
    assert result == 1, captured
    assert '/example.py\n' in captured
    assert ' func\n' in captured
    assert ' failed post-condition' in captured


def test_correct(tmp_path: Path):
    text = """
        import deal

        @deal.post(lambda x: x > 0)
        def func(x: int):
            return 1
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = main(['prove', str(path), '--nocolor'], stream=stream)

    stream.seek(0)
    captured = stream.read()
    assert result == 0, captured
    assert '/example.py\n' in captured
    assert ' func\n' in captured
    assert ' proved! post-condition' in captured


def test_skip(tmp_path: Path):
    text = """
        import deal

        def func(x):
            assert x > 0
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    result = main(['prove', str(path), '--nocolor'])
    assert result == 0


def test_skip_show_skipped(tmp_path: Path):
    text = """
        import deal

        def func(x):
            assert x > 0
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = main(['prove', str(path), '--nocolor', '--skipped'], stream=stream)

    stream.seek(0)
    captured = stream.read()
    assert result == 0, captured
    assert '/example.py\n' in captured
    assert ' func\n' in captured
    exp = 'skipped failed to interpret function arguments (missed annotation for x)\n'
    assert exp in captured


def test_skip_tests(tmp_path: Path):
    text = """
        import deal

        def test_func():
            assert False
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    result = main(['prove', str(path), '--nocolor'])
    assert result == 0


def test_prove_no_decorators_fail(tmp_path: Path):
    text = """
        import deal

        def func():
            assert False
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    result = main(['prove', str(path), '--nocolor'])
    assert result == 1


def test_prove_no_decorators_ok(tmp_path: Path):
    text = """
        import deal

        def func():
            assert True
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    result = main(['prove', str(path), '--nocolor'])
    assert result == 0


def test_show_filename_once(tmp_path: Path):
    text = """
        import deal

        def func1():
            assert True
        def func2():
            assert True
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = main(['prove', str(path), '--nocolor'], stream=stream)
    stream.seek(0)
    captured = stream.read()
    assert result == 0, captured
    assert 'func1' in captured
    assert 'func2' in captured
    assert captured.count('example.py') == 1



from typing import List

import deal


@deal.pre(lambda items: len(items) > 0)
@deal.has()
def my_min(items: List[int]) -> int:
    return min(items)


@deal.has('stdout')
def example():
    # good
    print(my_min([3, 1, 4]))
    # bad
    print(my_min([]))


test_min = deal.cases(my_min)



from ._contracts import Contracts
from ._decorators import (
    catch, chain, dispatch, ensure, example, has, implies,
    inherit, inv, post, pre, pure, raises, reason, safe,
)
from ._dispatch import Dispatch
from ._has_patcher import HasPatcher
from ._inherit import Inherit
from ._invariant import invariant
from ._validators import InvariantValidator, RaisesValidator, ReasonValidator, Validator


__all__ = [
    # public decorators
    'chain',
    'dispatch',
    'ensure',
    'example',
    'has',
    'inherit',
    'inv',
    'post',
    'pre',
    'raises',
    'reason',

    # public functions
    'catch',
    'implies',
    'pure',
    'safe',

    # private
    'Contracts',
    'Dispatch',
    'HasPatcher',
    'Inherit',
    'invariant',
    'InvariantValidator',
    'RaisesValidator',
    'ReasonValidator',
    'Validator',
]



import ast

import astroid
import pytest

from deal.linter._extractors import get_returns, has_returns


@pytest.mark.parametrize('text, expected', [
    ('return 1', (1, )),
    ('return -1', (-1, )),
    ('return 3.14', (3.14, )),
    ('return +3.14', (3.14, )),     # ignore unary plus
    ('return -3.14', (-3.14, )),    # handle unary minus
    ('return +a', ()),              # ignore uninferrable value inside of unary op
    ('return "lol"', ('lol', )),
    ('return b"lol"', (b'lol', )),
    ('return True', (True, )),
    ('return None', (None, )),
    ('return [1,2,3]', ([1, 2, 3], )),
    ('return', (None, )),

    ('if True: return 13', (13, )),
    ('if True:\n  return 13\nelse:\n  return 16', (13, 16)),
    ('for i in lst: return 13', (13, )),
    ('try:\n lol()\nexcept:\n 1\nelse:\n return 3', (3, )),
    ('try:\n lol()\nexcept:\n 1\nfinally:\n return 3', (3, )),
    ('with lol():\n return 3', (3, )),

    ('yield 1', (1, )),
    ('yield -1', (-1, )),
    ('yield True', (True, )),
    ('yield a', ()),
    ('yield', (None, )),
])
def test_get_returns_simple(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    returns = tuple(r.value for r in get_returns(body=tree.body))
    assert returns == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    returns = tuple(r.value for r in get_returns(body=tree.body))
    assert returns == expected


def test_ast_uninferrable_unary():
    tree = ast.parse('return ~4')
    print(ast.dump(tree))
    returns = tuple(r.value for r in get_returns(body=tree.body))
    assert returns == ()


@pytest.mark.parametrize('text, expected', [
    ('return 1 + 2', (3, )),                # do a simple arithmetic
    ('return a', ()),                       # ignore uninferrable names
    ('return ~4', (-5, )),                  # handle unary bitwise NOT

    ('a = 4\nreturn a', (4, )),             # infer variables
    ('a = str\nreturn a', ()),              # ignore not constants

    ('a = 4\nreturn [1, a]', ([1, 4], )),   # infer variables in list
    ('return [1, a]', ()),                  # ignore uninferralbe in list

    ('a = 4\nreturn (1, a)', ((1, 4), )),   # infer variables in set
    ('return (1, a)', ()),                  # ignore uninferralbe in set

    ('a = 4\nreturn {1, a}', ({1, 4}, )),   # infer variables in tuple
    ('return {1, a}', ()),                  # ignore uninferralbe in tuple
])
def test_get_returns_inference(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    returns = tuple(r.value for r in get_returns(body=tree.body))
    assert returns == expected


@pytest.mark.parametrize('text, expected', [
    ('return', True),
    ('return 1', True),
    ('if b:\n  return 1', True),
    ('yield 1', True),
    ('if b:\n  yield 1', True),
    ('1 + 2', False),
])
def test_has_returns(text, expected):
    tree = ast.parse(text)
    print(ast.dump(tree))
    assert has_returns(body=tree.body) is expected

    tree = astroid.parse(text)
    print(tree.repr_tree())
    assert has_returns(body=tree.body) is expected



import os

import pytest

import deal
import deal.introspection
from deal._imports import deactivate
from deal._state import state

from .test_runtime.helpers import run_sync


@pytest.fixture
def restore_state():
    state.reset()
    yield
    state.removed = False
    state.reset()
    deactivate()


def count_contracts(func) -> int:
    return len(list(deal.introspection.get_contracts(func)))


def test_contract_state_switch_custom_param(restore_state):
    func = deal.pre(lambda x: x > 0)(lambda x: x * 2)
    deal.disable()
    func(-2)
    deal.enable()
    with pytest.raises(deal.PreContractError):
        func(-2)


def test_contract_state_switch_default_param(restore_state):
    func = deal.pre(lambda x: x > 0)(lambda x: x * 2)
    deal.disable()
    assert func(-2) == -4
    deal.enable()
    with pytest.raises(deal.PreContractError):
        func(-2)


def test_contract_state_switch_default_param_async(restore_state):
    @deal.pre(lambda x: x > 0)
    async def func(x):
        return x * 2

    deal.disable()
    assert run_sync(func(-2)) == -4
    deal.enable()
    with pytest.raises(deal.PreContractError):
        run_sync(func(-2))


def test_contract_state_switch_default_param_generator(restore_state):
    @deal.pre(lambda x: x > 0)
    def func(x):
        yield x * 2

    deal.disable()
    assert list(func(-2)) == [-4]
    deal.enable()
    with pytest.raises(deal.PreContractError):
        list(func(-2))


def test_state_disable_permament(restore_state):
    @deal.pre(lambda x: x > 0)
    @deal.inherit
    @deal.pure
    def func1(x):
        yield x * 2

    deal.disable(permament=True)

    @deal.pre(lambda x: x > 0)
    @deal.inherit
    @deal.pure
    def func2(x):
        yield x * 2

    assert count_contracts(func1) == 3
    assert count_contracts(func2) == 0


def test_state_disable_permament__cant_disable_twice(restore_state):
    deal.disable(permament=True)
    with pytest.raises(RuntimeError):
        deal.disable(permament=True)
    with pytest.raises(RuntimeError):
        deal.enable()
    with pytest.raises(RuntimeError):
        deal.reset()


def test_state_switch_module_load(restore_state):
    with pytest.raises(RuntimeError):
        deal.module_load()
    deal.disable()
    deal.activate()
    deal.module_load()


def test_state_switch_module_load_debug(restore_state):
    with pytest.raises(RuntimeError):
        deal.module_load()
    deal.disable()
    deal.activate()
    deal.enable()


def test_state_switch_activate(restore_state):
    assert deal.activate()
    assert deactivate()

    deal.disable()
    assert not deal.activate()


@pytest.fixture
def set_env_vars():
    old_vars = os.environ.copy()
    yield os.environ.update
    os.environ.clear()
    os.environ.update(old_vars)


@pytest.mark.parametrize('env_vars, expected', [
    (dict(), None),
    (dict(CI='true'), None),
    (dict(GCLOUD_PROJECT='example'), 'It is GCP but deal is enabled'),
    (dict(LAMBDA_TASK_ROOT='/home/'), 'It is AWS but deal is enabled'),
])
def test_enable__warnings(restore_state, env_vars, set_env_vars, expected):
    os.environ.clear()
    set_env_vars(env_vars)
    ewarn = RuntimeWarning if expected else None
    with pytest.warns(ewarn) as warns:
        deal.enable()
    if expected:
        assert len(warns) == 1
        assert str(warns[0].message) == f'{expected}. Is it intentional?'
    else:
        assert len(warns) == 0

    with pytest.warns(None) as warns:
        deal.enable(warn=False)
    assert len(warns) == 0


@pytest.mark.parametrize('env_vars, expected', [
    (dict(), None),
    (dict(GCLOUD_PROJECT='example'), None),
    (dict(LAMBDA_TASK_ROOT='/home/'), None),
    (dict(CI='true'), 'It is CI but deal is disabled'),
    (dict(PYTEST_CURRENT_TEST='test_example'), 'It is pytest but deal is disabled'),
])
def test_disable__warnings(restore_state, env_vars, set_env_vars, expected):
    os.environ.clear()
    set_env_vars(env_vars)
    ewarn = RuntimeWarning if expected else None
    with pytest.warns(ewarn) as warns:
        deal.disable()
    if expected:
        assert len(warns) == 1
        assert str(warns[0].message) == f'{expected}. Is it intentional?'
    else:
        assert len(warns) == 0

    with pytest.warns(None) as warns:
        deal.disable(warn=False)
    assert len(warns) == 0



from typing import List

import hypothesis

import deal


@deal.pre(lambda items: len(items) > 0)
@deal.has()
def my_min(items: List[int]) -> int:
    return min(items)


@hypothesis.example([1, 2, 3])
@deal.cases(
    func=my_min,
    settings=hypothesis.settings(
        verbosity=hypothesis.Verbosity.normal,
    ),
)
def test_min(case):
    case()


if __name__ == '__main__':
    test_min()



from __future__ import annotations

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TextIO


class Command:
    stream: TextIO
    root: Path

    def __init__(self, stream: TextIO, root: Path) -> None:
        self.stream = stream
        self.root = root

    def print(self, *args) -> None:
        print(*args, file=self.stream)

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        raise NotImplementedError

    def __call__(self, args: Namespace) -> int:
        raise NotImplementedError



from __future__ import annotations

from functools import partial
from typing import Callable, TypeVar, Union, overload

from .. import _exceptions
from .._types import ExceptionType
from ._contracts import Contracts
from ._dispatch import Dispatch
from ._has_patcher import HasPatcher
from ._inherit import Inherit
from ._invariant import invariant
from ._validators import InvariantValidator, RaisesValidator, ReasonValidator, Validator


C = TypeVar('C', bound=Callable)
F = TypeVar('F', bound=Callable)
T = TypeVar('T')
TF = TypeVar('TF', bound=Union[Callable, type])


def pre(
    validator,
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """Decorator implementing precondition [value] contract.

    [Precondition][wiki] is
    a condition that must be true before the function is executed.
    Raises `PreContractError` otherwise.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        ``PreContractError`` by default.
    :return: a function wrapper.

    ```pycon
    >>> import deal
    >>> @deal.pre(lambda a, b: a + b > 0)
    ... def example(a, b):
    ...     return (a + b) * 2
    >>> example(1, 2)
    6
    >>> example(1, -2)
    Traceback (most recent call last):
      ...
    PreContractError: expected a + b > 0 (where a=1, b=-2)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Precondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    contract = Validator(
        validator=validator,
        message=message,
        exception=exception or _exceptions.PreContractError,
    )
    func = partial(Contracts.attach, 'pres', contract)
    return func  # type: ignore[return-value]


def post(
    validator,
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """Decorator implementing postcondition [value] contract.

    [Postcondition][wiki] is
    a condition that must be true for the function result.
    Raises `PostContractError` otherwise.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        ``PostContractError`` by default.
    :return: a function wrapper.

    ```pycon
    >>> import deal
    >>> @deal.post(lambda res: res > 0)
    ... def example(a, b):
    ...     return a + b
    >>> example(-1, 2)
    1
    >>> example(1, -2)
    Traceback (most recent call last):
      ...
    PostContractError: expected res > 0 (where res=-1)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Postcondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    contract = Validator(
        validator=validator,
        message=message,
        exception=exception or _exceptions.PostContractError,
    )
    func = partial(Contracts.attach, 'posts', contract)
    return func  # type: ignore[return-value]


def ensure(
    validator,
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """Decorator implementing postcondition [value] contract.

    [Postcondition][wiki] is
    a condition that must be true for the function result.
    Raises `PostContractError` otherwise.
    It's like [@deal.post](#deal.post) but contract accepts as input value
    not only the function result but also the function input arguments.
    The function result is passed into validator as `result` keyword argument.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        `PostContractError` by default.
    :return: a function wrapper.

    ```pycon
    >>> import deal
    >>> @deal.ensure(lambda a, result: a < result)
    ... def example(a):
    ...     return a * 2
    >>> example(2)
    4
    >>> example(0)
    Traceback (most recent call last):
      ...
    PostContractError: expected a < result (where result=0, a=0)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Postcondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    contract = Validator(
        validator=validator,
        message=message,
        exception=exception or _exceptions.PostContractError,
    )
    func = partial(Contracts.attach, 'ensures', contract)
    return func  # type: ignore[return-value]


def raises(
    *exceptions: type[Exception],
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """Decorator listing the exceptions which the function can raise.

    Implements [exception] contract.
    If the function raises an exception not listed in the decorator,
    `RaisesContractError` will be raised.

    :param exceptions: exceptions which the function can raise.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        `RaisesContractError` by default.
    :return: a function wrapper.

    ```pycon
    >>> import deal
    >>> @deal.raises(ZeroDivisionError, ValueError)
    ... def div(a, b):
    ...   return a / b
    ...
    >>> div(1, 0)
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    >>> div(1, '')
    Traceback (most recent call last):
        ...
     TypeError: unsupported operand type(s) for /: 'int' and 'str'
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError

    ```

    [exception]: https://deal.readthedocs.io/basic/exceptions.html
    """
    contract = RaisesValidator(
        exceptions=exceptions,
        message=message,
        exception=exception or _exceptions.RaisesContractError,
    )
    func = partial(Contracts.attach, 'raises', contract)
    return func  # type: ignore[return-value]


def has(
    *markers: str,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """Decorator controlling [side-effects] of the function.

    Allows to specify markers identifying which side-effect the functon may have.
    Side-effects must be propagated into all callers using `deal.has`,
    this is controlled by the [linter].
    Some side-effects are also checked at runtime.


    ```pycon
    >>> import deal
    >>> @deal.has()
    ... def greet():
    ...   print('hello')
    ...
    >>> greet()
    Traceback (most recent call last):
        ...
    SilentContractError
    >>> @deal.has('stdout')
    ... def greet():
    ...   print('hello')
    ...
    >>> greet()
    hello

    ```

    [side-effects]: https://deal.readthedocs.io/basic/side-effects.html
    [linter]: https://deal.readthedocs.io/basic/linter.html
    """
    patcher = HasPatcher(
        markers=markers,
        message=message,
        exception=exception,
    )
    func = partial(Contracts.attach_has, patcher)
    return func  # type: ignore[return-value]


def reason(
    event: type[Exception],
    validator,
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    """
    Decorator implementing [exception] contract.

    Allows to assert precondition for raised exception.
    It's like [@deal.ensure](#deal.ensure) but when instead of returning result
    the function raises an exception.

    :param event: exception raising which will trigger contract validation.
    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        `ReasonContractError` by default.
    :return: a function wrapper.

    ```pycon
    >>> import deal
    >>> @deal.reason(ZeroDivisionError, lambda a, b: b == 0)
    ... def div(a, b):
    ...   return a / (a - b)
    >>> div(2, 1)
    2.0
    >>> div(0, 0)
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    >>> div(2, 2)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    ReasonContractError: expected b == 0 (where a=2, b=2)

    ```

    [exception]: https://deal.readthedocs.io/basic/exceptions.html
    """
    contract = ReasonValidator(
        event=event,
        validator=validator,
        message=message,
        exception=exception or _exceptions.ReasonContractError,
    )
    func = partial(Contracts.attach, 'reasons', contract)
    return func  # type: ignore[return-value]


def inv(
    validator,
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[T], T]:
    """
    Decorator implementing invariant [value] contract.

    [Invariant][wiki] is a condition that can be relied upon to be true during execution
    of a program. `@deal.inv` is triggered in 3 cases:

    1. Before class method execution.
    2. After class method execution.
    3. After some class attribute setting.

    Deal doesn't rollback changes on contract violation.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :param exception: exception type to raise on the contract violation.
        `InvContractError` by default.
    :return: a class wrapper.

    ```pycon
    >>> import deal
    >>> @deal.inv(lambda obj: obj.likes >= 0)
    ... class Video:
    ...   likes = 1
    ...   def like(self): self.likes += 1
    ...   def dislike(self): self.likes -= 1
    ...
    >>> v = Video()
    >>> v.dislike()
    >>> v.likes
    0
    >>> v.dislike()
    Traceback (most recent call last):
    ...
    InvContractError: expected obj.likes >= 0
    >>> v.likes
    -1
    >>> v.likes = 2
    >>> v.likes = -2
    Traceback (most recent call last):
    ...
    InvContractError: expected obj.likes >= 0
    >>> v.likes
    -2

    ```

    [wiki]: https://en.wikipedia.org/wiki/Class_invariant
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    contract = InvariantValidator(
        validator=validator,
        message=message,
        exception=exception or _exceptions.InvContractError,
    )
    return partial(invariant, contract)


def example(validator: Callable[[], bool]) -> Callable[[C], C]:
    """
    Decorator for providing a usage example for the wrapped function.

    The example isn't checked at runtime.
    Instead, it is run in tests and checked by the linter.
    The example should use the decorated function
    and return True if the result is expected.

    ```pycon
    >>> import deal
    >>> @deal.example(lambda: double(3) == 6)
    ... def double(x):
    ...   return x * 2
    ...

    ```
    """
    contract = Validator(
        validator=validator,
        message=None,
        exception=_exceptions.ExampleContractError,
    )
    func = partial(Contracts.attach, 'examples', contract)
    return func  # type: ignore[return-value]


@overload
def safe(
    *,
    message: str | None = None,
    exception: ExceptionType | None = None,
) -> Callable[[C], C]:
    pass


@overload
def safe(_func: C) -> C:
    pass


def safe(_func=None, **kwargs):
    """
    Alias for [@deal.raises()](#deal.raises).
    Wraps a function that never raises an exception.

    ```pycon
    >>> import deal
    >>> @deal.safe
    ... def div(a, b):
    ...   return a / b
    ...
    >>> div(2, 4)
    0.5
    >>> div(2, 0)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError

    ```
    """
    if _func is None:
        return raises(**kwargs)
    return raises()(_func)


def chain(*contracts: Callable[[C], C]) -> Callable[[F], F]:
    """Decorator to chain 2 or more contracts together.

    It can be helpful to store contracts separately from the function.
    Consider using it when you have too many contracts.
    Otherwise, the function will be lost under a bunch of decorators.

    ```pycon
    >>> import deal
    >>> sum_contract = deal.chain(
    ...   deal.pre(lambda a, b: a > 0),
    ...   deal.pre(lambda a, b: b > 0),
    ...   deal.post(lambda res: res > 0),
    ... )
    >>> @sum_contract
    ... def sum(a, b):
    ...   return a + b
    ...
    >>> sum(2, 3)
    5
    >>> sum(2, -3)
    Traceback (most recent call last):
        ...
    PreContractError: expected b > 0 (where a=2, b=-3)
    >>> sum(-2, 3)
    Traceback (most recent call last):
        ...
    PreContractError: expected a > 0 (where a=-2, b=3)

    ```

    :param contracts: contracts to chain.
    :return: a function wrapper
    """
    def wrapped(func):
        for contract in contracts:
            func = contract(func)
        return func
    return wrapped


def pure(_func: C) -> C:
    """Decorator for [pure functions][wiki].

    Alias for `@deal.chain(deal.has(), deal.safe)`.

    Pure function has no side-effects and doesn't raise any exceptions.

    ```pycon
    >>> import deal
    >>> @deal.pure
    ... def div(a, b, log=False):
    ...   if log:
    ...     print('div called')
    ...   return a / b
    ...
    >>> div(2, 4)
    0.5
    >>> div(2, 0)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError
    >>> div(2, 3, log=True)
    Traceback (most recent call last):
        ...
    SilentContractError

    ```

    [wiki]: https://en.wikipedia.org/wiki/Pure_function
    """
    return chain(has(), safe)(_func)


def implies(test, then: T) -> Union[bool, T]:
    """Check `then` only if `test` is true.

    A convenient helper for contracts that must be checked only for some cases.
    It is known as "implication" or [material conditional][wiki].

    ```pycon
    >>> import deal
    >>> deal.implies(False, False)
    True
    >>> deal.implies(False, True)
    True
    >>> deal.implies(True, False)
    False
    >>> deal.implies(True, True)
    True

    ```

    [wiki]: https://en.wikipedia.org/wiki/Material_conditional
    """
    return not test or then


def catch(func: Callable, *args, **kwargs) -> type[Exception] | None:
    """Call the function with the given arguments, catching any exception.

    The catched exception is returned.
    This function may be useful in combination with {py:func}`deal.example`.

    ```pycon
    >>> import deal
    >>> @deal.example(lambda: deal.catch(div, 4, 0) is ZeroDivisionError)
    ... @deal.raises(ZeroDivisionError)
    ... @deal.reason(ZeroDivisionError, lambda x: x == 0)
    ... def div(x, y):
    ...   return x / y
    ...
    >>>

    ```
    """
    try:
        func(*args, **kwargs)
    except Exception as exc:
        return type(exc)
    return None


def dispatch(func: C) -> Dispatch[C]:
    """Combine multiple functions into one.

    When the decorated function is called, it will try to call all registered
    functions and return the result from the first one that doesn't raise
    `PreContractError`.

    ```pycon
    >>> import deal
    >>> @deal.dispatch
    ... def double(x: int) -> int:
    ...   raise NotImplementedError
    ...
    >>> @double.register
    ... @deal.pre(lambda x: x == 3)
    ... def _(x: int) -> int:
    ...   return 6
    ...
    >>> @double.register
    ... @deal.pre(lambda x: x == 4)
    ... def _(x: int) -> int:
    ...   return 8
    ...
    >>> double(3)
    6
    >>> double(4)
    8
    >>> double(5)
    Traceback (most recent call last):
        ...
    NoMatchError: expected x == 3 (where x=5); expected x == 4 (where x=5)

    ```

    """
    return Dispatch.wrap(func)


def inherit(func: TF) -> TF:
    """Inherit contracts from base classes.

    Can be used to decorate either the whole class or a separate method.

    ```pycon
    >>> import deal
    >>> class Shape:
    ...   @deal.post(lambda r: r > 2)
    ...   def get_sides(self):
    ...     raise NotImplementedError
    ...
    >>> class Triangle(Shape):
    ...   @deal.inherit
    ...   def get_sides(self):
    ...     return 3
    ...
    >>> class Line(Shape):
    ...   @deal.inherit
    ...   def get_sides(self):
    ...     return 2
    ...
    >>> triangle = Triangle()
    >>> triangle.get_sides()
    3
    >>> line = Line()
    >>> line.get_sides()
    Traceback (most recent call last):
        ...
    PreContractError: expected r > 0 (where r=2)

    ```
    """
    return Inherit.wrap(func)



import ast

import astroid
import pytest

from deal.linter._extractors.common import _get_module, get_full_name, get_name, infer


@pytest.mark.parametrize('text, expected', [
    ('name', 'name'),
    ('left.right', 'left.right'),

    ('left().right', None),
    ('1', None),
])
def test_get_name(text, expected):
    tree = ast.parse(text)
    print(ast.dump(tree))
    expr = tree.body[0].value
    assert get_name(expr=expr) == expected

    tree = astroid.parse(text)
    print(tree.repr_tree())
    expr = tree.body[-1].value
    assert get_name(expr=expr) == expected


@pytest.mark.parametrize('text, expected', [
    ('unknown', []),
    ('from pathlib import Path\nPath.write_text', [('pathlib', 'Path.write_text')]),
    ('def f(): pass\nf', [('', 'f')]),
    ('class C:\n def f(): pass\nC.f', [('', 'C.f')]),
    ('class C:\n def f(): pass\nc = C()\nc.f', [('', 'C.f')]),
    ('print', [('builtins', 'print')]),
    ('"".format', [('builtins', 'str.format')]),
])
def test_infer(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    expr = tree.body[-1].value
    actual = infer(expr=expr)
    assert [get_full_name(e) for e in actual] == expected


def test_get_full_name_func():
    tree = astroid.parse('def f(): pass')
    print(tree.repr_tree())
    func = tree.body[0]
    assert get_full_name(expr=func) == ('', 'f')


def test_get_full_name_method():
    tree = astroid.parse('class C:\n  def f(): pass')
    print(tree.repr_tree())
    func = tree.body[0].body[0]
    assert get_full_name(expr=func) == ('', 'C.f')


def test_get_full_name_deep_method():
    tree = astroid.parse('class A:\n  class B:\n    def f(): pass')
    print(tree.repr_tree())
    func = tree.body[0].body[0].body[0]
    assert get_full_name(expr=func) == ('', 'A.B.f')


def test_get_full_name_func_in_func():
    tree = astroid.parse('def outer():\n def inner(): pass')
    print(tree.repr_tree())
    func = tree.body[0].body[0]
    assert get_full_name(expr=func) == ('', 'outer.inner')


def test_get_full_name_not_a_func():
    tree = astroid.parse('try:\n pass\nexcept E as e:\n def f(): pass')
    print(tree.repr_tree())
    func = tree.body[0].handlers[0].body[0]
    assert get_full_name(expr=func) == ('', 'f')


def test_get_full_name_no_parent():
    tree = astroid.parse('def f(): pass')
    print(tree.repr_tree())
    assert get_full_name(expr=tree) == ('', '')


def test_get_module():
    tree = astroid.parse('def f(): pass')
    print(tree.repr_tree())
    assert _get_module(expr=tree) is tree
    assert _get_module(expr=tree.body[0]) is tree

    tree.body[0].parent = None
    assert _get_module(expr=tree.body[0]) is None



from __future__ import annotations

from typing import Any

import vaa


class Scheme:
    """A base class for implementing a custom validator.

    The custom validator should implement `is_valid` method.
    The method should return True if the data is valid.
    Otherwise, it should set `errors` attribute and return False.
    """
    data: dict[str, Any]
    errors: list[vaa.Error]

    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def is_valid(self) -> bool:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Callable, Generic, TypeVar, overload


T = TypeVar('T')


class cached_property(Generic[T]):  # noqa: N801
    def __init__(self, func: Callable[[Any], T]) -> None:
        self.func = func

    @overload
    def __get__(self, instance: None, owner: type | None = ...) -> cached_property[T]:
        pass

    @overload
    def __get__(self, instance, owner: type | None = ...) -> T:
        pass

    def __get__(self, obj, cls):
        value = self.func(obj)
        obj.__dict__[self.func.__name__] = value
        return value



from __future__ import annotations

from pathlib import Path
from typing import Iterator


def get_paths(path: Path) -> Iterator[Path]:
    """Recursively yields python files.
    """
    if not path.exists():
        raise FileNotFoundError(str(path))
    if path.is_file():
        if path.suffix == '.py':
            yield path
        return
    for subpath in path.iterdir():
        if subpath.name[0] == '.':
            continue
        if subpath.name == '__pycache__':
            continue
        yield from get_paths(subpath)



from pathlib import Path
from textwrap import dedent

import pytest
from sphinx.cmd.build import build_main

import deal


contracts = deal.chain(
    deal.reason(ZeroDivisionError, lambda a, b: b == 0),
    deal.reason(ValueError, lambda a, b: a == b, message='a is equal to b'),
    deal.raises(ValueError, IndexError, ZeroDivisionError),
    deal.post(lambda res: res != .13),
    deal.ensure(lambda a, b, result: b != result),
    deal.pre(lambda a, b: b != 0, message='b is not zero'),
    deal.pre(lambda a, b: b != 0),
    deal.has('database'),
)


@contracts
@deal.example(lambda: example_sphinx(6, 2) == 3)
def example_sphinx(a: int, b: int) -> float:
    """Example function.

    :return: The description for return value.
    """
    return a / b


@contracts
@deal.example(lambda: example_google(6, 2) == 3)
def example_google(a: int, b: int) -> float:
    """Example function.

    Returns:
        The description for return value.
    """
    return a / b


def example_plain(a: int, b: int) -> float:
    """Example function.

    Returns:
        The description for return value.
    """
    return a / b


CONFIG = """
    import deal

    extensions = [
        'sphinx.ext.autodoc',
        'sphinx.ext.napoleon',
    ]

    def setup(app):
        deal.autodoc(app)
"""


@pytest.mark.parametrize('style', ['sphinx', 'google'])
def test_autodoc_smoke(style: str, tmp_path: Path):
    path_in = tmp_path / 'in'
    path_in.mkdir()
    path_out = tmp_path / 'out'
    (path_in / 'conf.py').write_text(dedent(CONFIG))
    (path_in / 'index.rst').write_text(dedent(f"""
        .. autofunction:: tests.test_sphinx.example_{style}
    """))
    exit_code = build_main([str(path_in), str(path_out), '-b', 'text', '-ET'])
    assert exit_code == 0
    content = (path_out / 'index.txt').read_text()
    assert 'ValueError' in content
    assert 'ZeroDivisionError' in content
    assert 'b == 0' in content

    expected = dedent(f"""
        tests.test_sphinx.example_{style}(a: int, b: int) -> float
            Example function.
            Returns:
                The description for return value.
            Side-effects:
                * database
            Raises:
                * **IndexError** --
                * **ValueError** -- a is equal to b
                * **ZeroDivisionError** -- "b == 0"
            Contracts:
                * b is not zero
                * "b != 0"
                * "res != .13"
                * "b != result"
            Examples:
                * "example_{style}(6, 2) == 3"
    """)

    lines = [line.strip() for line in content.splitlines()]
    content = '\n'.join(line for line in lines if line)

    lines = [line.strip() for line in expected.splitlines()]
    expected = '\n'.join(line for line in lines if line)

    assert content.strip() == expected.strip()


def test_autodoc_no_contracts(tmp_path: Path):
    path_in = tmp_path / 'in'
    path_in.mkdir()
    path_out = tmp_path / 'out'
    (path_in / 'conf.py').write_text(dedent(CONFIG))
    (path_in / 'index.rst').write_text(
        '.. autofunction:: tests.test_sphinx.example_plain',
    )
    exit_code = build_main([str(path_in), str(path_out), '-b', 'text', '-ET'])
    assert exit_code == 0
    content = (path_out / 'index.txt').read_text()
    assert 'tests.test_sphinx.example_plain' in content
    assert 'Contracts' not in content



from __future__ import annotations

from functools import update_wrapper
from inspect import iscoroutinefunction, isgeneratorfunction
from typing import TYPE_CHECKING, Callable, Generic, TypeVar

from .._exceptions import ContractError
from .._state import state


if TYPE_CHECKING:
    from ._has_patcher import HasPatcher
    from ._validators import RaisesValidator, ReasonValidator, Validator


F = TypeVar('F', bound=Callable)
ATTR = '__deal_contract'


class Contracts(Generic[F]):
    __slots__ = (
        'func',
        'wrapped',
        'pres',
        'posts',
        'ensures',
        'examples',
        'raises',
        'reasons',
        'patcher',
    )

    func: F
    wrapped: F
    pres: list[Validator]
    posts: list[Validator]
    ensures: list[Validator]
    examples: list[Validator]
    raises: list[RaisesValidator]
    reasons: list[ReasonValidator]
    patcher: HasPatcher | None

    def __init__(self, func: F) -> None:
        self.func = func
        self.pres = []
        self.posts = []
        self.ensures = []
        self.examples = []
        self.raises = []
        self.reasons = []
        self.patcher = None

    @classmethod
    def attach(cls, contract_type: str, validator: Validator, func: F) -> F:
        if state.removed:
            return func
        contracts = cls._ensure_wrapped(func)
        validator.function = func
        getattr(contracts, contract_type).append(validator)
        return contracts.wrapped

    @classmethod
    def attach_has(cls, patcher: HasPatcher, func: F) -> F:
        if state.removed:
            return func
        contracts = cls._ensure_wrapped(func)
        contracts.patcher = patcher
        return contracts.wrapped

    @classmethod
    def _ensure_wrapped(cls: type[Contracts], func: F) -> Contracts[F]:
        contracts: Contracts
        contracts = getattr(func, ATTR, None)  # type: ignore[assignment]
        if contracts is not None:
            return contracts
        contracts = cls(func)
        assert contracts is not None

        if iscoroutinefunction(func):
            async def wrapper(*args, **kwargs):
                return await contracts._run_async(args, kwargs)
        elif isgeneratorfunction(func):
            def wrapper(*args, **kwargs):
                yield from contracts._run_iter(args, kwargs)
        else:
            def wrapper(*args, **kwargs):
                return contracts._run_sync(args, kwargs)

        update_wrapper(wrapper=wrapper, wrapped=func)
        setattr(wrapper, ATTR, contracts)
        contracts.wrapped = wrapper
        return contracts

    def wrap(self, func: F) -> F:
        """Wrap the function with the contracts.
        """
        contracts = self._ensure_wrapped(func)
        contracts.pres.extend(self.pres)
        contracts.posts.extend(self.posts)
        contracts.ensures.extend(self.ensures)
        contracts.examples.extend(self.examples)
        contracts.raises.extend(self.raises)
        contracts.reasons.extend(self.reasons)
        if self.patcher is not None:
            if contracts.patcher is not None:
                contracts.patcher.markers |= self.patcher.markers
            else:
                contracts.patcher = self.patcher
        return contracts.wrapped

    def _run_sync(self, args: tuple[object, ...], kwargs: dict[str, object]):
        if not state.debug:
            return self.func(*args, **kwargs)

        # pre-validation
        state.debug = False
        try:
            for validator in self.pres:
                validator.validate(args, kwargs)
        finally:
            state.debug = True

        # running the function
        if self.patcher is not None:
            self.patcher.patch()
        try:
            result = self.func(*args, **kwargs)
        except ContractError:
            raise
        except Exception as exc:
            for validator in self.raises:
                validator.validate(args, kwargs, exc=exc)
            exc_type = type(exc)
            for validator in self.reasons:
                if exc_type is validator.event:
                    validator.validate(args, kwargs, exc=exc)
            raise
        finally:
            if self.patcher is not None:
                self.patcher.unpatch()

        # post-validation
        state.debug = False
        try:
            for validator in self.posts:
                validator.validate((result,), {})
            for validator in self.ensures:
                validator.validate(args, dict(kwargs, result=result))
        finally:
            state.debug = True

        return result

    async def _run_async(self, args: tuple[object, ...], kwargs: dict[str, object]):
        if not state.debug:
            return await self.func(*args, **kwargs)

        # pre-validation
        state.debug = False
        try:
            for validator in self.pres:
                validator.validate(args, kwargs)
        finally:
            state.debug = True

        # running the function
        if self.patcher is not None:
            self.patcher.patch()
        try:
            result = await self.func(*args, **kwargs)
        except ContractError:
            raise
        except Exception as exc:
            for validator in self.raises:
                validator.validate(args, kwargs, exc=exc)
            exc_type = type(exc)
            for validator in self.reasons:
                if exc_type is validator.event:
                    validator.validate(args, kwargs, exc=exc)
            raise
        finally:
            if self.patcher is not None:
                self.patcher.unpatch()

        # post-validation
        state.debug = False
        try:
            for validator in self.posts:
                validator.validate((result,), {})
            for validator in self.ensures:
                validator.validate(args, dict(kwargs, result=result))
        finally:
            state.debug = True

        return result

    def _run_iter(self, args: tuple[object, ...], kwargs: dict[str, object]):
        if not state.debug:
            yield from self.func(*args, **kwargs)
            return

        # pre-validation
        state.debug = False
        try:
            for validator in self.pres:
                validator.validate(args, kwargs)
        finally:
            state.debug = True

        generator = self.func(*args, **kwargs)
        while True:
            # running the function
            if self.patcher is not None:
                self.patcher.patch()
            try:
                result = next(generator)
            except StopIteration:
                return
            except ContractError:
                raise
            except Exception as exc:
                for validator in self.raises:
                    validator.validate(args, kwargs, exc=exc)
                exc_type = type(exc)
                for validator in self.reasons:
                    if exc_type is validator.event:
                        validator.validate(args, kwargs, exc=exc)
                raise
            finally:
                if self.patcher is not None:
                    self.patcher.unpatch()

            # post-validation
            state.debug = False
            try:
                for validator in self.posts:
                    validator.validate((result,), {})
                for validator in self.ensures:
                    validator.validate(args, dict(kwargs, result=result))
            finally:
                state.debug = True
            yield result



import ast

import astroid
import pytest

from deal.linter._extractors import get_definitions


@pytest.mark.parametrize('source, names', [
    ('import re', {'re'}),
    ('import typing, types', {'typing', 'types'}),
    ('import typing as types', {'types'}),
    ('from . import hi', set()),
    ('from .something import hi', set()),

    ('from typing import List', {'List'}),
    ('from typing import List, Dict', {'List', 'Dict'}),

    ('ab = 2', {'ab'}),
    ('ab = cd = 23', {'ab', 'cd'}),
    ('ab.cd = 2', set()),
])
def test_extract_defs(source: str, names) -> None:
    tree = ast.parse(source)
    print(ast.dump(tree))
    defs = get_definitions(tree)
    assert set(defs) == names

    tree = astroid.parse(source)
    print(tree.repr_tree())
    defs = get_definitions(tree)
    assert set(defs) == names



import pytest
import urllib3

import deal


def test_pure_silent():
    @deal.pure
    def func(msg):
        if msg:
            print(msg)

    func(None)
    with pytest.raises(deal.SilentContractError):
        func('bad')


def test_pure_safe():
    func = deal.pure(lambda x: 1 / x)
    func(2)
    with pytest.raises(deal.RaisesContractError):
        func(0)


def test_pure_offline():
    @deal.pure
    def func(do):
        if do:
            http = urllib3.PoolManager()
            http.request('GET', 'http://httpbin.org/robots.txt')

    func(False)
    with pytest.raises(deal.OfflineContractError):
        func(True)



import ast
from textwrap import dedent

import astroid

from deal.linter._func import Func


TEXT = """
    import deal

    @deal.post(lambda x: x > 0)
    @deal.raises(ValueError, UnknownError, idk_what_is_it())
    @notadeal.raises(KeyError)
    @deal.notaraises(KeyError)
    @something
    @another.deco
    @something()
    def f(x):
        return x

    def h(x):
        return x

    @something()
    def k(x):
        return x
"""


def test_from_text():
    funcs = Func.from_text(dedent(TEXT))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_ast():
    funcs = Func.from_ast(ast.parse(dedent(TEXT)))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_astroid():
    funcs = Func.from_astroid(astroid.parse(dedent(TEXT)))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_ast_methods():
    text = dedent("""
        class A:
            p = 1
            @deal.post(lambda x: x > 0)
            def f(self, x):
                return x
    """)
    funcs = Func.from_ast(ast.parse(text))
    assert len(funcs) == 1
    assert len(funcs[0].contracts) == 1


def test_from_astroid_methods():
    text = dedent("""
        class A:
            p = 1
            @deal.post(lambda x: x > 0)
            def f(self, x):
                return x
    """)
    funcs = Func.from_astroid(astroid.parse(text))
    assert len(funcs) == 1
    assert len(funcs[0].contracts) == 1


def test_repr():
    funcs1 = Func.from_ast(ast.parse(dedent(TEXT)))
    funcs2 = Func.from_astroid(astroid.parse(dedent(TEXT)))
    for func in (funcs1[0], funcs2[0]):
        assert repr(func) == 'Func(post, raises)'



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Iterator, NamedTuple, Sequence

import astroid

from ._contract import Category


EXTENSION = '.json'
ROOT = Path(__file__).parent / 'stubs'
CPYTHON_ROOT = ROOT / 'cpython'


class StubFile:
    __slots__ = ('path', '_content')
    path: Path
    _content: dict[str, dict[str, Any]]

    def __init__(self, path: Path) -> None:
        self.path = path
        self._content = dict()

    def load(self) -> None:
        with self.path.open(encoding='utf8') as stream:
            self._content = json.load(stream)

    def dump(self) -> None:
        if not self._content:
            return
        with self.path.open(mode='w', encoding='utf8') as stream:
            json.dump(obj=self._content, fp=stream, indent=2, sort_keys=True)

    def add(self, func: str, contract: Category, value: str) -> None:
        if contract not in (Category.RAISES, Category.HAS):
            raise ValueError('unsupported contract')
        contracts = self._content.setdefault(func, dict())
        values = contracts.setdefault(contract.value, [])
        if value in values:
            return
        values.append(value)
        values.sort()

    def get(self, func: str, contract: Category) -> frozenset[str]:
        if contract not in (Category.RAISES, Category.HAS):
            raise ValueError('unsupported contract')
        values = self._content.get(func, {}).get(contract.value, [])
        return frozenset(values)


class StubsManager:
    __slots__ = ('paths', '_modules')
    _modules: dict[str, StubFile]
    paths: tuple[Path, ...]

    default_paths = (ROOT, CPYTHON_ROOT)

    def __init__(self, paths: Sequence[Path] | None = None) -> None:
        self._modules = dict()
        if paths is None:
            self.paths = self.default_paths
        else:
            self.paths = tuple(paths)

    def read(self, *, path: Path, module_name: str | None = None) -> StubFile:
        if path.suffix == '.py':
            path = path.with_suffix(EXTENSION)
        if path.suffix != EXTENSION:
            raise ValueError(f'invalid stub file extension: *{path.suffix}')
        if module_name is None:
            module_name = self._get_module_name(path=path)
        if module_name not in self._modules:
            stub = StubFile(path=path)
            stub.load()
            self._modules[module_name] = stub
        return self._modules[module_name]

    @staticmethod
    def _get_module_name(path: Path) -> str:
        path = path.resolve()
        # walk up by the tree as pytest does
        if not (path.parent / '__init__.py').exists():
            return path.stem
        for parent in path.parents:
            if not (parent / '__init__.py').exists():
                parts = path.relative_to(parent).with_suffix('').parts
                return '.'.join(parts)
        raise RuntimeError('unreachable: __init__.py files up to root?')  # pragma: no cover

    def get(self, module_name: str) -> StubFile | None:
        # cached
        stub = self._modules.get(module_name)
        if stub is not None:
            return stub
        # in the root
        for root in self.paths:
            path = root / (module_name + EXTENSION)
            if path.exists():
                return self.read(path=path, module_name=module_name)
            path = root.joinpath(*module_name.split('.')).with_suffix(EXTENSION)
            if path.exists():
                return self.read(path=path, module_name=module_name)
        return None

    def create(self, path: Path) -> StubFile:
        if path.suffix == '.py':
            path = path.with_suffix(EXTENSION)
        module_name = self._get_module_name(path=path)

        # if the stub for file is somewhere in the paths, use this instead.
        stub = self.get(module_name=module_name)
        if stub is not None:
            return stub

        # create new stub and load it from disk if the file exists
        stub = StubFile(path=path)
        if path.exists():
            stub.load()
        self._modules[module_name] = stub
        return stub


class PseudoFunc(NamedTuple):
    name: str
    body: list


def _get_funcs(*, path: Path) -> Iterator[PseudoFunc]:
    text = path.read_text()
    tree = astroid.parse(code=text, path=str(path))
    for expr in tree.body:
        yield from _get_funcs_from_expr(expr=expr)


def _get_funcs_from_expr(expr, prefix: str = '') -> Iterator[PseudoFunc]:
    name = getattr(expr, 'name', '')
    if prefix:
        name = prefix + '.' + name

    # functions
    if type(expr) is astroid.FunctionDef:
        yield PseudoFunc(name=name, body=expr.body)
    if type(expr) is astroid.AsyncFunctionDef:
        yield PseudoFunc(name=name, body=expr.body)

    # methods
    if type(expr) is astroid.ClassDef:
        for subexpr in expr.body:
            yield from _get_funcs_from_expr(expr=subexpr, prefix=name)


def generate_stub(*, path: Path, stubs: StubsManager | None = None) -> Path:
    from ._extractors import get_exceptions, get_markers

    if path.suffix != '.py':
        raise ValueError(f'invalid Python file extension: *{path.suffix}')

    if stubs is None:
        stubs = StubsManager()
    stub = stubs.create(path=path)
    for func in _get_funcs(path=path):
        for token in get_exceptions(body=func.body, stubs=stubs):
            value = token.value
            if isinstance(value, type):
                value = value.__name__
            stub.add(func=func.name, contract=Category.RAISES, value=str(value))
        for token in get_markers(body=func.body, stubs=stubs):
            assert token.marker is not None
            stub.add(func=func.name, contract=Category.HAS, value=token.marker)
    stub.dump()
    return stub.path



import sys

import pytest

import deal

from .helpers import run_sync


def test_not_allow_print():
    @deal.has()
    def func(msg):
        if msg:
            print(msg)

    func(None)
    with pytest.raises(deal.SilentContractError):
        func('bad')


def test_allow_print():
    @deal.has('stdout')
    def func(msg):
        if msg:
            print(msg)

    func(None)
    func('good')


def test_not_allow_stderr():
    @deal.has()
    def func(msg):
        if msg:
            print(msg, file=sys.stderr)

    func(None)
    with pytest.raises(deal.SilentContractError):
        func('bad')


def test_allow_stderr():
    @deal.has('stderr')
    def func(msg):
        if msg:
            print(msg, file=sys.stderr)

    func(None)
    func('good')


def test_decorating_async_function():
    @deal.has()
    async def func(msg):
        if msg:
            print(msg)
        return msg

    assert run_sync(func('')) == ''
    with pytest.raises(deal.SilentContractError):
        run_sync(func('a'))


def test_decorating_generator():
    @deal.has()
    def func(msg):
        if msg:
            print(msg)
        yield msg

    assert list(func('')) == ['']
    with pytest.raises(deal.SilentContractError):
        list(func('a'))



import deal


def preserve_function_annotation() -> None:
    @deal.dispatch
    def double(a: int) -> int:
        raise NotImplementedError

    @double.register
    @deal.pre(lambda a: a == 2)
    def _1(a: int) -> int:
        return 4

    # Preserve function type
    # R: deal._runtime._dispatch.Dispatch[def (a: builtins.int) -> builtins.int]
    reveal_type(double)

    res = double(2)
    reveal_type(res)  # R: builtins.int

    # Enforce contrvariance
    # E: Argument 1 to "register" of "Dispatch" has incompatible type "Callable[[bool], int]"; expected "Callable[[int], int]"
    @double.register
    def _2(a: bool) -> int:
        return 4

    # Enforce covariance
    # E: Argument 1 to "register" of "Dispatch" has incompatible type "Callable[[int], float]"; expected "Callable[[int], int]"
    @double.register
    def _3(a: int) -> float:
        return True



from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

from .._colors import get_colors
from ..linter import TransformationType, Transformer
from ._base import Command
from ._common import get_paths


class DecorateCommand(Command):
    """Add decorators to your code.

    ```bash
    python3 -m deal decorate project/
    ```

    Options:
    + `--types`: types of decorators to apply. All are enabled by default.
    + `--double-quotes`: use double quotes. Single quotes are used by default.
    + `--nocolor`: do not use colors in the console output.

    The exit code is always 0. If you want to test the code for missed decorators,
    use the `lint` command instead.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            '--types',
            nargs='*',
            choices=[tt.value for tt in TransformationType],
            default=sorted(tt.value for tt in TransformationType),
            help='types of decorators to apply',
        )
        parser.add_argument(
            '--double-quotes',
            action='store_true',
            help='use double quotes',
        )
        parser.add_argument('--nocolor', action='store_true', help='colorless output')
        parser.add_argument('paths', nargs='*', default='.')

    def __call__(self, args) -> int:
        types = {TransformationType(t) for t in args.types}
        colors = get_colors(args)
        for arg in args.paths:
            for path in get_paths(Path(arg)):
                self.print('{magenta}{path}{end}'.format(path=path, **colors))
                original_code = path.read_text(encoding='utf8')
                tr = Transformer(
                    content=original_code,
                    path=path,
                    types=types,
                )
                if args.double_quotes:
                    tr = tr._replace(quote='"')
                modified_code = tr.transform()
                if original_code == modified_code:
                    self.print('  {blue}no changes{end}'.format(**colors))
                else:
                    path.write_text(modified_code)
                    self.print('  {green}decorated{end}'.format(**colors))
        return 0



import deal


def preserve_function_annotation() -> None:
    @deal.post(lambda r: r > 0)
    def f() -> int:
        return 4
    reveal_type(f)  # R: def () -> builtins.int


def infer_validator_type() -> None:
    # E: "int" has no attribute "hi"
    @deal.post(lambda r: r.hi)
    def f() -> int:
        return 4


def report_bad_arity() -> None:
    # TODO: Fix "Cannot infer type of lambda"

    # E: Argument 1 to "post" has incompatible type "Callable[[], bool]"; expected "Callable[[int], Union[bool, str]]"
    @deal.post(lambda: True)  # E: Cannot infer type of lambda
    def f() -> int:
        return 4


def detect_methods() -> None:
    class A:
        # E: "int" has no attribute "hi"
        @deal.post(lambda r: r.hi)
        def f(self) -> int:
            return 4


def do_not_fail_on_vaa_simple() -> None:
    @deal.post(lambda _: _.result == 4)
    def f(self) -> int:
        return 4


def report_bad_return_type() -> None:
    # E: Argument 1 to "post" has incompatible type "Callable[[int], int]"; expected "Callable[[int], Union[bool, str]]"
    @deal.post(lambda x: x)  # E: Incompatible return value type (got "int", expected "Union[bool, str]")
    def f(x) -> int:
        return 4



from __future__ import annotations

import sys
from contextlib import suppress
from pathlib import Path
from typing import Any

from ._cached_property import cached_property
from ._colors import COLORS, NOCOLORS, highlight
from ._source import get_validator_source
from ._state import state


root = str(Path(__file__).parent)


def exception_hook(etype: type[BaseException], value: BaseException, tb):
    """Exception hook to remove deal from the traceback for ContractError.
    """
    if not issubclass(etype, ContractError):
        return _excepthook(etype, value, tb)

    # try to reduce traceback by removing `deal` part
    prev_tb = patched_tb = tb
    while patched_tb:
        path: str = patched_tb.tb_frame.f_code.co_filename
        if path.startswith(root):
            with suppress(AttributeError):  # read-only attribute in <3.7
                prev_tb.tb_next = None
            break  # pragma: no cover
        prev_tb = patched_tb
        patched_tb = patched_tb.tb_next
    else:
        # cannot find deal in the trace, leave it as is
        patched_tb = tb

    return _excepthook(etype, value, patched_tb)


_excepthook = sys.excepthook
sys.excepthook = exception_hook


class ContractError(AssertionError):
    """The base class for all errors raised by deal contracts.
    """
    message: str
    errors: Any | None
    validator: Any
    params: dict[str, Any]
    origin: object | None

    def __init__(
        self,
        message: str = '',
        errors=None,
        validator=None,
        params: dict[str, Any] | None = None,
        origin: object | None = None,
    ) -> None:
        self.message = message
        self.errors = errors
        self.validator = validator
        self.params = params or {}
        self.origin = origin

        args = []
        if message:
            args.append(message)
        if errors:
            args.append(errors)
        super().__init__(*args)

    @cached_property
    def source(self) -> str:
        """The raw unformatted source code of the validator.
        """
        if self.validator is None:
            return ''
        source = get_validator_source(self.validator)
        if source:
            return source
        if hasattr(self.validator, '__name__'):
            return self.validator.__name__
        return repr(self.validator)

    @cached_property
    def colored_source(self) -> str:
        """The colored source code of the validator.
        """
        return highlight(self.source)

    @cached_property
    def variables(self) -> str:
        """Formatted variables passed into the validator.
        """
        sep = ', '
        colors = COLORS
        if not state.color:
            colors = NOCOLORS
        tmpl = '{blue}{k}{end}={magenta}{v}{end}'
        params = []
        for k, v in self.params.items():
            v = repr(v)
            if len(v) > 20:
                continue
            params.append(tmpl.format(k=k, v=v, **colors))
        return sep.join(params)

    def __str__(self) -> str:
        result = self.message
        if not result and self.errors:
            result = repr(self.errors)
        if not result and self.source:
            result = 'expected '
            if state.color:
                result += self.colored_source
            else:
                result += self.source
        if self.variables:
            result += f' (where {self.variables})'
        return result


class PreContractError(ContractError):
    """The error raised by `deal.pre` for contract violation.
    """


class PostContractError(ContractError):
    """The error raised by `deal.post` for contract violation.
    """


class InvContractError(ContractError):
    """The error raised by `deal.inv` for contract violation.
    """


class ExampleContractError(ContractError):
    """The error raised by `deal.example` for contract violation.

    `deal.example` contracts are checked only during testing and linting,
    not at runtime.
    """


class RaisesContractError(ContractError):
    """The error raised by `deal.raises` for contract violation.
    """


class ReasonContractError(ContractError):
    """The error raised by `deal.reason` for contract violation.
    """


class MarkerError(ContractError):
    """The base class for errors raised by `deal.has` for contract violation.
    """


class OfflineContractError(MarkerError):
    """The error raised by `deal.has` for networking markers violation.

    The networking can be allowed by markers `io`, `network`, and `socket`.
    """


class SilentContractError(MarkerError):
    """The error raised by `deal.has` for printing markers violation.

    The printing can be allowed by markers `io`, `print`, `stdout`, and `stderr`.
    """


class NoMatchError(Exception):
    """The error raised by `deal.dispatch` when there is no matching implementation.

    "No matching implementation" means that all registered functions
    raised `PreContractError`.
    """
    __module__ = 'deal'

    def __init__(self, exceptions: tuple[PreContractError, ...]) -> None:
        self.exceptions = exceptions

    def __str__(self) -> str:
        return '; '.join(str(e) for e in self.exceptions)


# Patch module name to show in repr `deal` instead of `deal._exceptions`
for cls in ContractError.__subclasses__():
    cls.__module__ = 'deal'
for cls in MarkerError.__subclasses__():
    cls.__module__ = 'deal'



import deal


def preserve_function_annotation() -> None:
    @deal.pre(lambda a, b: a > b)
    def f(a: int, b: int) -> str:
        return str(a + b)
    reveal_type(f)  # R: def (a: builtins.int, b: builtins.int) -> builtins.str


def infer_validator_type() -> None:
    # E: "int" has no attribute "hi"
    @deal.pre(lambda a, b: a.hi)
    def f(a: int, b: int) -> bool:
        return a > b



from __future__ import annotations

import ast
from typing import Union

import astroid

from .common import TOKENS, Extractor, Token, traverse
from .value import UNKNOWN, get_value


get_returns = Extractor()


def has_returns(body: list) -> bool:
    expected = TOKENS.RETURN + TOKENS.YIELD + TOKENS.YIELD_FROM + TOKENS.RAISE
    for expr in traverse(body=body):
        if isinstance(expr, expected):
            return True
    return False


@get_returns.register(*TOKENS.RETURN)
def handle_return(expr: Union[ast.Return, astroid.Return]) -> Token | None:
    if expr.value is None:
        return Token(value=None, line=expr.lineno, col=expr.col_offset)
    value = get_value(expr=expr.value)
    if value is UNKNOWN:
        return None
    return Token(value=value, line=expr.lineno, col=expr.value.col_offset)


@get_returns.register(*TOKENS.YIELD)
def handle_yield(expr: Union[ast.Yield, astroid.Yield]) -> Token | None:
    if expr.value is None:
        return Token(value=None, line=expr.lineno, col=expr.col_offset)
    value = get_value(expr=expr.value)
    if value is UNKNOWN:
        return None
    return Token(value=value, line=expr.lineno, col=expr.value.col_offset)



from __future__ import annotations

import ast
from types import MappingProxyType
from typing import Iterator, Type, TypeVar

import astroid

from .._runtime import HasPatcher
from ._contract import Category, Contract
from ._error import Error
from ._extractors import (
    UNKNOWN, get_asserts, get_example, get_exceptions, get_imports,
    get_markers, get_pre, get_returns, get_value, has_returns, uses_result,
)
from ._func import Func
from ._stub import StubsManager


T = TypeVar('T', bound=Type['Rule'])
rules: list[Rule] = []


def register(rule: T) -> T:
    rules.append(rule())
    return rule


class Rule:
    __slots__ = ()
    code: int
    message: str

    def _validate(self, contract: Contract, args, kwargs, **error_info) -> Error | None:
        try:
            result = contract.run(*args, **kwargs)
        except NameError:
            # cannot resolve contract dependencies
            return None
        if isinstance(result, str):
            return Error(text=result, code=self.code, **error_info)
        if not result:
            return Error(text=self.message, code=self.code, **error_info)
        return None


class ModuleRule(Rule):
    __slots__ = ()

    def __call__(self, tree: ast.Module) -> Iterator[Error]:
        raise NotImplementedError


class FuncRule(Rule):
    __slots__ = ()

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        raise NotImplementedError


@register
class CheckImports(ModuleRule):
    __slots__ = ()
    code = 1
    message = 'do not use `from deal import ...`, use `import deal` instead'

    def __call__(self, tree: ast.Module) -> Iterator[Error]:
        for token in get_imports(tree.body):
            if token.value != 'deal':
                continue
            yield Error(
                code=self.code,
                text=self.message,
                row=token.line,
                col=token.col,
            )


@register
class CheckEnsureArgs(FuncRule):
    __slots__ = ()
    code = 2
    message = 'ensure contract must have `result` arg'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        for contract in func.contracts:
            if contract.category != Category.ENSURE:
                continue
            yield from self._check(contract)

    def _check(self, contract: Contract) -> Iterator[Error]:
        validator = contract.args[0]
        if not uses_result(validator):
            yield Error(
                code=self.code,
                text=self.message,
                row=validator.lineno,
                col=validator.col_offset,
            )


@register
class CheckPre(FuncRule):
    __slots__ = ()
    code = 11
    message = 'pre contract error'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        # We test only contracted functions because of poor performance.
        # Inferring every called function in the whole project
        # is a really expensive operation.
        if not func.contracts:
            return
        context = func.contracts[0].context
        for token in get_pre(body=func.body, context=context):
            yield Error(
                code=self.code,
                text=token.marker or self.message,
                value=token.value,  # type: ignore
                row=token.line,
                col=token.col,
            )


@register
class CheckReturns(FuncRule):
    __slots__ = ()
    code = 12
    message = 'post contract error'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        for contract in func.contracts:
            if contract.category != Category.POST:
                continue
            yield from self._check(func=func, contract=contract)

    def _check(self, func: Func, contract: Contract) -> Iterator[Error]:
        for token in get_returns(body=func.body):
            error = self._validate(
                contract=contract,
                args=(token.value,),
                kwargs={},
                row=token.line,
                col=token.col,
                value=str(token.value),
            )
            if error is not None:
                yield error


@register
class CheckExamples(FuncRule):
    __slots__ = ()
    code = 13
    message = 'example violates contract'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        for contract in func.contracts:
            if contract.category != Category.EXAMPLE:
                continue
            yield from self._check(func=func, contract=contract)

    def _check(self, func: Func, contract: Contract) -> Iterator[Error]:
        token = contract.args[0]
        if not isinstance(token, (ast.Lambda, astroid.Lambda)):
            return
        example = get_example(token.body, func_name=func.name)
        if example is None:
            return
        for other in func.contracts:
            if other.category == Category.PRE:
                error = self._validate(
                    contract=other,
                    args=example.args,
                    kwargs=example.kwargs,
                    row=token.lineno,
                    col=token.col_offset,
                    value='deal.pre',
                )
                if error is not None:
                    yield error

            if other.category == Category.POST:
                if example.result == UNKNOWN:
                    continue
                error = self._validate(
                    contract=other,
                    args=(example.result, ),
                    kwargs={},
                    row=token.lineno,
                    col=token.col_offset,
                    value='deal.post',
                )
                if error is not None:
                    yield error

            if other.category == Category.ENSURE:
                if example.result == UNKNOWN:
                    continue
                error = self._validate(
                    contract=other,
                    args=example.args,
                    kwargs=dict(example.kwargs, result=example.result),
                    row=token.lineno,
                    col=token.col_offset,
                    value='deal.ensure',
                )
                if error is not None:
                    yield error


@register
class CheckRaises(FuncRule):
    __slots__ = ()
    code = 21
    message = 'raises contract error'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        cats = {Category.RAISES, Category.SAFE, Category.PURE}
        declared: list[str | type[Exception]] = [AssertionError]
        check = False
        for contract in func.contracts:
            if contract.category not in cats:
                continue
            declared.extend(contract.exceptions)
            check = True
        if check:
            yield from self.get_undeclared(func, declared, stubs)

    def get_undeclared(
        self,
        func: Func,
        declared: list[str | type[Exception]],
        stubs: StubsManager | None = None,
    ) -> Iterator[Error]:
        declared_types = tuple(exc for exc in declared if not isinstance(exc, str))
        for token in get_exceptions(body=func.body, stubs=stubs):
            if token.value in declared:
                continue
            exc = token.value
            if isinstance(exc, type):
                if issubclass(exc, declared_types):
                    continue
                exc = exc.__name__
            yield Error(
                code=self.code,
                text=self.message,
                value=str(exc),
                row=token.line,
                col=token.col,
            )


@register
class CheckAsserts(FuncRule):
    __slots__ = ()
    code = 31
    message = 'assert error'

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        # do not validate asserts in tests
        if func.name.startswith('test_'):
            return
        for token in get_asserts(body=func.body):
            yield Error(
                code=self.code,
                text=self.message,
                value=str(token.value),
                row=token.line,
                col=token.col,
            )


@register
class CheckMarkers(FuncRule):
    __slots__ = ()
    code = 40
    message = 'missed marker'

    codes = MappingProxyType({
        'global': 41,
        'import': 42,

        'io': 43,
        'read': 44,
        'write': 45,
        'stdout': 46,
        'stderr': 47,
        'network': 48,
        'stdin': 49,
        'syscall': 50,

        'random': 55,
        'time': 56,
    })

    def __call__(self, func: Func, stubs: StubsManager | None = None) -> Iterator[Error]:
        for contract in func.contracts:
            markers: set[str] | None = None
            if contract.category == Category.HAS:
                markers = set()
                for arg in contract.args:
                    value = get_value(arg)
                    if isinstance(value, str):
                        markers.add(value)
            elif contract.category == Category.PURE:
                markers = set()
            if markers is None:  # this is needed to fix coverage quirks
                continue
            yield from self.get_undeclared(func=func, markers=markers)
            return

    @classmethod
    def get_undeclared(cls, func: Func, markers: set[str]) -> Iterator[Error]:
        has = HasPatcher(markers)
        # function without IO must return something
        if not has.has_io and not func.has_self and not has_returns(body=func.body):
            yield Error(
                code=cls.codes['io'],
                text=cls.message,
                value='io',
                row=func.line,
                col=func.col,
            )

        for token in get_markers(body=func.body):
            assert token.marker
            has_marker = getattr(has, f'has_{token.marker}', None)
            if has_marker is None:
                has_marker = token.marker in has.markers
            if has_marker:
                continue
            yield Error(
                code=cls.codes.get(token.marker, cls.code),
                text=cls.message,
                value=token.marker,
                row=token.line,
                col=token.col,
            )



from .asserts import get_asserts
from .common import get_name
from .contracts import get_contracts
from .definitions import get_definitions
from .examples import get_example
from .exceptions import get_exceptions
from .imports import get_imports
from .markers import get_markers
from .pre import get_pre
from .result import uses_result
from .returns import get_returns, has_returns
from .value import UNKNOWN, get_value


__all__ = [
    'get_asserts',
    'get_contracts',
    'get_definitions',
    'get_example',
    'get_exceptions',
    'get_imports',
    'get_markers',
    'get_name',
    'get_pre',
    'get_returns',
    'get_value',
    'has_returns',
    'uses_result',
    'UNKNOWN',
]



import pytest

import deal


def test_usage_example():
    source = deal.introspection.__doc__.split('```python')[1].split('```')[0]
    # deal cannot extract the source code for validator if there is no source file
    source = source.replace("'x > 0'", "''")
    exec(source)


def test_get_contracts__pre():
    def validator(x):
        return x > 0

    def func(x):
        return x * 2

    wrapped = deal.pre(validator)(func)
    contracts = list(deal.introspection.get_contracts(wrapped))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Pre
    assert isinstance(contract, deal.introspection.Pre)
    contract.validate(2)
    with pytest.raises(deal.PreContractError):
        contract.validate(-1)
    assert contract.exception is deal.PreContractError
    assert contract.message is None
    assert contract.source == 'validator'


def test_unwrap():
    def validator(x):
        return x > 0

    def func(x):
        return x * 2

    wrapped = deal.pre(validator)(func)
    assert deal.introspection.unwrap(wrapped) is func
    assert deal.introspection.unwrap(func) is func


def test_custom_exception_and_message():
    @deal.pre(lambda x: x > 0, exception=ValueError, message='msg')
    def func(x):
        return x * 2

    contract = next(deal.introspection.get_contracts(func))
    assert isinstance(contract, deal.introspection.Pre)
    assert type(contract.exception) is ValueError
    assert contract.message == 'msg'


def test_get_contracts__raises():
    @deal.raises(ZeroDivisionError)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Raises
    assert isinstance(contract, deal.introspection.Raises)
    assert contract.exceptions == (ZeroDivisionError,)


def test_get_contracts__reason():
    @deal.reason(ZeroDivisionError, lambda x: x > 0)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Reason
    assert isinstance(contract, deal.introspection.Reason)
    assert contract.event == ZeroDivisionError


def test_get_contracts__has():
    @deal.has('io', 'db')
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Has
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'io', 'db'})
    assert contract.exception is deal.MarkerError
    assert contract.exception_type is deal.MarkerError
    assert contract.message is None


def test_get_contracts__multiple():
    @deal.pre(lambda x: x > 0)
    @deal.post(lambda x: x >= 2)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 2
    assert type(contracts[0]) is deal.introspection.Pre
    assert type(contracts[1]) is deal.introspection.Post


def test_get_contracts__example():
    @deal.example(lambda: func(3) == 6)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Example
    assert isinstance(contract, deal.introspection.Example)
    contract.validate()
    assert contract.exception is deal.ExampleContractError
    assert contract.message is None
    assert contract.source == 'func(3) == 6'


def test_get_contracts__example_func():
    from examples.sphinx import example

    contracts = list(deal.introspection.get_contracts(example))
    assert len(contracts) == 9


def test_init():
    @deal.pre(lambda x: x > 0)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 1
    contract = contracts[0]
    assert type(contract) is deal.introspection.Pre
    val = contract._wrapped
    assert val.validate.__name__ == '_init'
    contract.init()
    assert val.validate.__name__ != '_init'
    assert val.signature


def test_init_all():
    @deal.pre(lambda x: x > 0)
    @deal.pure
    @deal.post(lambda x: x > 0)
    def func(x):
        return x * 2

    contracts = list(deal.introspection.get_contracts(func))
    assert len(contracts) == 4
    val1 = contracts[0]._wrapped
    val2 = contracts[1]._wrapped
    assert val1.validate.__name__ == '_init'
    assert val2.validate.__name__ == '_init'
    deal.introspection.init_all(func)
    assert val1.validate.__name__ != '_init'
    assert val2.validate.__name__ != '_init'
    assert val1.signature
    assert val2.signature


def test_get_contracts__inherit_class():
    class A:
        @deal.has('a', 'b')
        def f(self, x):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        def f(self, x):
            return x * 2

    b = B()
    contracts = list(deal.introspection.get_contracts(b.f))
    assert len(contracts) == 1
    contract = contracts[0]
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'a', 'b'})


def test_get_contracts__inherit_func():
    class A:
        @deal.has('a', 'b')
        def f(self, x):
            raise NotImplementedError

    class B(A):
        @deal.inherit
        def f(self, x):
            return x * 2

    b = B()
    contracts = list(deal.introspection.get_contracts(b.f))
    assert len(contracts) == 1
    contract = contracts[0]
    assert isinstance(contract, deal.introspection.Has)
    assert contract.markers == frozenset({'a', 'b'})



from __future__ import annotations

import socket
import sys
from io import StringIO
from typing import Iterable

from .._exceptions import MarkerError, OfflineContractError, SilentContractError
from .._types import ExceptionType


KNOWN_MARKERS = frozenset({
    # io markers
    'io',
    'network',
    'read',
    'stderr',
    'stdin',
    'stdout',
    'syscall',
    'write',

    # non-io markers
    'global',
    'import',
    'random',
    'time',

    # aliases
    'input',     # stdin
    'nonlocal',  # global
    'print',     # stdout
    'socket',    # network
})
NON_IO_MARKERS = frozenset({
    'global',
    'nonlocal',
    'import',
    'random',
    'time',
})


class PatchedStringIO(StringIO):
    __slots__ = ('exception',)

    def __init__(self, exception: ExceptionType) -> None:
        self.exception = exception

    def write(self, *args, **kwargs):
        raise self.exception


class PatchedSocket:
    __slots__ = ('exception',)

    def __init__(self, exception: ExceptionType) -> None:
        self.exception = exception

    def __call__(self, *args, **kwargs):
        raise self.exception


class HasPatcher:
    __slots__ = (
        'markers',
        'message',
        'exception',
        'true_socket',
        'true_stdout',
        'true_stderr',
    )
    markers: frozenset[str]

    def __init__(
        self,
        markers: Iterable[str],
        message: str | None = None,
        exception: ExceptionType | None = None,
    ) -> None:
        self.markers = frozenset(markers)
        self.message = message
        self.exception = exception or MarkerError
        if message and isinstance(self.exception, type):
            self.exception = self.exception(message)

    @property
    def exception_type(self) -> type[Exception]:
        if isinstance(self.exception, Exception):
            return type(self.exception)
        return self.exception

    @property
    def has_network(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'network' in self.markers:
            return True
        if 'socket' in self.markers:
            return True
        return False

    @property
    def has_io(self) -> bool:
        return bool(self.markers - NON_IO_MARKERS)

    @property
    def has_stdout(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'print' in self.markers:
            return True
        if 'stdout' in self.markers:
            return True
        return False

    @property
    def has_stderr(self) -> bool:
        if 'io' in self.markers:
            return True
        return 'stderr' in self.markers

    @property
    def has_global(self) -> bool:
        if 'global' in self.markers:
            return True
        if 'nonlocal' in self.markers:
            return True
        return False

    @property
    def has_read(self) -> bool:
        if 'io' in self.markers:
            return True
        return 'read' in self.markers

    @property
    def has_stdin(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'input' in self.markers:
            return True
        if 'stdin' in self.markers:
            return True
        return False

    @property
    def has_write(self) -> bool:
        if 'io' in self.markers:
            return True
        return 'write' in self.markers

    # patching

    def patch(self) -> None:
        if not self.has_network:
            self.true_socket = socket.socket
            socket.socket = PatchedSocket(  # type: ignore[assignment,misc]
                exception=self._get_exception(OfflineContractError),
            )
        if not self.has_stdout:
            self.true_stdout = sys.stdout
            sys.stdout = PatchedStringIO(
                exception=self._get_exception(SilentContractError),
            )
        if not self.has_stderr:
            self.true_stderr = sys.stderr
            sys.stderr = PatchedStringIO(
                exception=self._get_exception(SilentContractError),
            )

    def unpatch(self) -> None:
        if not self.has_network:
            socket.socket = self.true_socket  # type: ignore[misc]
        if not self.has_stdout:
            sys.stdout = self.true_stdout
        if not self.has_stderr:
            sys.stderr = self.true_stderr

    def _get_exception(self, default: type[Exception]) -> ExceptionType:
        if self.exception_type is MarkerError:
            if self.message is None:
                return default
            return default(self.message)
        return self.exception
